// 20.09.2017, Arduino v1.8.4, LZ2WSG, KN34PC
// Генератор с два бутона за настройка на стъпката, енкодер и/или два бутона за настройка на честотата,
// запомняне на последно използваната стъпка и честота във външен EEPROM 24C02
//---------------------------------------------------------------------------------------------------------
#include <Wire.h>
#include <LiquidCrystal.h>
#include "Rotary.h"                  // http://www.buxtronix.net/2011/10/rotary-encoders-done-properly.html
#include "at24c02.h"

#define FQ_UD           10
#define W_CLK           9
#define DATA            8
#define F_MCLK          179999000    // 180000000 14.05.2019

#define SW_LEFT         11
#define SW_RIGHT        A1
#define SW_DOWN         A2
#define SW_UP           A3

#define SW_STEP         A0

#define DEFAULT_FREQ    10000000
#define DEFAULT_INC     1000
#define F_MIN           1
#define F_MAX           99999999

#define EEPROM_I2C_ADDR 0x50                     // EEPROM 24C02 I2C адрес
#define EEPROM_SECRET   0x73                     // 'таен' байт, адрес 0x00
#define EEPROM_T_MS     10000                    // време до запис в ext EEPROM 

#define pulse_high(pin) {digitalWrite(pin, HIGH); digitalWrite(pin, LOW);}

at24c02 eeprom(EEPROM_I2C_ADDR);
LiquidCrystal lcd(12, 13, 7, 6, 5, 4);           // RS(4), E(6), D4(11), D5(12), D6(13), D7(14)
Rotary r = Rotary(2, 3);

int32_t rx;
int32_t rx_serial;
int32_t enc_inc;
int8_t enc_dir = 0;                              // -1 DIR_CCW, 0 DIR_NONE, 1 DIR_CCW
bool ch_flag = true;
uint8_t inc_num;
bool rep = false;

bool eeprom_found = false;
bool eeprom_error = false;
bool eeprom_flag = true;
bool eeprom_freq_flag = true;
bool eeprom_inc_flag = true;
uint32_t eeprom_t = 0;
//---------------------------------------------------------------------------------------------------------
void setup() {
  Serial.begin(57600);
  if (Serial) {
    Serial.setTimeout(200);
    Serial.println("Enter Fo [Hz]:");
  }

  lcd.begin(16, 2);
  lcd.setCursor(0, 0);
  lcd.print("Fo = ");
  lcd.setCursor(14, 0);
  lcd.print("Hz");

  pinMode(FQ_UD, OUTPUT);
  pinMode(W_CLK, OUTPUT);
  pinMode(DATA, OUTPUT);

  pinMode(SW_LEFT, INPUT_PULLUP);
  pinMode(SW_RIGHT, INPUT_PULLUP);
  pinMode(SW_UP, INPUT_PULLUP);
  pinMode(SW_DOWN, INPUT_PULLUP);

  pinMode(SW_STEP, INPUT_PULLUP);

  attachInterrupt(0, rotary_encoder, CHANGE);
  attachInterrupt(1, rotary_encoder, CHANGE);

  ad9851_init();
  eeprom.init();

  if (i2c_device_check(EEPROM_I2C_ADDR)) {       // има ли свързана по I2C ext EEPROM?
    eeprom_found = true;
    if (eeprom.read(0) == EEPROM_SECRET) {       // ext EEPROM е моя?
      inc_num = eeprom.read(1);
      if (inc_num > 8) {                         // клетката в ext EEPROM за inc_num е дефектна?
        enc_inc = DEFAULT_INC;
        eeprom_error = true;
      }
      else {
        enc_inc = 1;
        for (uint8_t d = 0; d < inc_num; d++)
          enc_inc *= 10;
      }

      rx = eeprom_read_freq(2);
      if ((rx < F_MIN) || (rx > F_MAX)) {        // клетките в ext EEPROM за rx са дефектирали?
        rx = DEFAULT_FREQ;
        eeprom_error = true;
      }

      if (eeprom_error)
        show_error_message();
    }
    else {                                       // ext EEPROM е нова, запис на данни по подразбиране
      eeprom.write(0, EEPROM_SECRET);

      inc_num = log10(DEFAULT_INC);
      enc_inc = DEFAULT_INC;
      eeprom.write(1, inc_num);

      rx = DEFAULT_FREQ;
      eeprom_write_freq(2, rx);
    }
  }
  else {                                         // няма ext EEPROM, използване на данни по подразбиране
    rx = DEFAULT_FREQ;
    enc_inc = DEFAULT_INC;
    eeprom_found = false;
  }
  show_inc();
}
//---------------------------------------------------------------------------------------------------------
void loop() {
  if (ch_flag) {
    if (digitalRead(SW_STEP)) {
      switch (enc_dir) {
        case -1: if ((rx - enc_inc) >= F_MIN) rx -= enc_inc; break;
        case  1: if ((rx + enc_inc) <= F_MAX) rx += enc_inc; break;
      }
      enc_dir = 0;
      ad9851_set_freq(rx);
      show_freq(rx);

      if (eeprom_found) {                        // ако има свързана на I2C ext EEPROM
        eeprom_flag = false;                     // има данни за записване
        eeprom_freq_flag = false;                // има нова честота за записване
        eeprom_t = millis();
      }
    }
    else
      do {
        switch (enc_dir) {
          case -1: enc_inc *= 10; break;
          case  1: enc_inc /= 10; break;
        }

        if (enc_dir != 0) {
          enc_inc = constrain(enc_inc, 1, 10000000);
          show_inc();
          if (eeprom_found) {                    // ако има свързана на I2C ext EEPROM
            eeprom_flag = false;                 // има данни за записване
            eeprom_inc_flag = false;             // има нова стъпка за записване
            eeprom_t = millis();
          }
          enc_dir = 0;
        }
      }
      while (!digitalRead(SW_STEP));

    ch_flag = false;
  }

  if (Serial) {                                  // ако серийния интерфейс е достъпен
    if (Serial.available() > 0) {
      String str = Serial.readString();
      rx_serial = str.toInt();
      switch (str[str.length() - 1]) {           // chars [k, K, m, M]
        case 'k':                                // 1 kHz = 1000 Hz
        case 'K': rx_serial *= 1000; break;      // 1 kHz = 1000 Hz
        case 'm':                                // 1 MHz = 1000000 Hz
        case 'M': rx_serial *= 1000000; break;   // 1 MHz = 1000000 Hz
      }
      if ((rx_serial < F_MIN) || (rx_serial > F_MAX))
        Serial.println("Enter output frequency in range [1 Hz - 99999999 Hz]");
      else {
        rx = rx_serial;
        Serial.print("Fo = ");
        Serial.print(rx);
        Serial.println(" Hz");
        ch_flag = true;
      }
    }
  }

  if (!digitalRead(SW_LEFT)) {
    if (button_state(SW_LEFT) > 0) {
      enc_inc *= 10;
      if (enc_inc > 10000000)
        enc_inc = 10000000;
      show_inc();
      if (eeprom_found) {                        // ако има свързана на I2C ext EEPROM
        eeprom_flag = false;                     // има данни за записване
        eeprom_inc_flag = false;                 // има нова стъпка за записване
        eeprom_t = millis();
      }
    }
  }

  if (!digitalRead(SW_RIGHT)) {
    if (button_state(SW_RIGHT) > 0) {
      enc_inc /= 10;
      if (enc_inc < 1)
        enc_inc = 1;
      show_inc();
      if (eeprom_found) {                        // ако има свързана на I2C ext EEPROM
        eeprom_flag = false;                     // има данни за записване
        eeprom_inc_flag = false;                 // има нова стъпка за записване
        eeprom_t = millis();
      }
    }
  }

  if (!digitalRead(SW_UP)) {
    if (button_state(SW_UP) > 0) {
      enc_dir = 1;
      ch_flag = true;
    }
  }

  if (!digitalRead(SW_DOWN)) {
    if (button_state(SW_DOWN) > 0) {
      enc_dir = -1;
      ch_flag = true;
    }
  }

  if ((digitalRead(SW_UP)) && (digitalRead(SW_DOWN)))
    rep = false;

  if (!eeprom_flag) {
    if ((eeprom_t + EEPROM_T_MS) < millis()) {   // след 10 sec запис на текущата честота в ext EEPROM
      if (!eeprom_inc_flag) {
        inc_num = log10(enc_inc);
        eeprom.update(1, inc_num);
        eeprom_inc_flag = true;
      }
      if (!eeprom_freq_flag) {
        eeprom_write_freq(2, rx);
        eeprom_freq_flag = true;
      }
      eeprom_flag = true;
    }
  }
}
//---------------------------------------------------------------------------------------------------------
void rotary_encoder() {                          // rotary encoder events
  uint8_t result = r.process();

  if (result) {
    ch_flag = true;
    if (result == DIR_CW)
      enc_dir = 1;
    else
      enc_dir = -1;
  }
}
//---------------------------------------------------------------------------------------------------------
void ad9851_init() {
  pulse_high(W_CLK);
  pulse_high(FQ_UD);
}
//---------------------------------------------------------------------------------------------------------
void ad9851_set_freq(double frequency) {
  uint32_t freq = frequency * 4294967296 / F_MCLK;

  noInterrupts();
  shiftOut(DATA, W_CLK, LSBFIRST, freq);         // freq_07 - freq_00
  shiftOut(DATA, W_CLK, LSBFIRST, (freq >> 8));  // freq_08 - freq_15
  shiftOut(DATA, W_CLK, LSBFIRST, (freq >> 16)); // freq_16 - freq_23
  shiftOut(DATA, W_CLK, LSBFIRST, (freq >> 24)); // freq_24 - freq_31
  shiftOut(DATA, W_CLK, LSBFIRST, 0b00000001);   // phase_4 - phase_0, power-down, 0, 6 x REFCLK Multiplier
  pulse_high(FQ_UD);
  interrupts();
}
//---------------------------------------------------------------------------------------------------------
void show_freq(uint32_t sf_rx) {
  lcd.setCursor(5, 0);
  if (sf_rx < 10000000)
    lcd.print('0');
  if (sf_rx < 1000000)
    lcd.print('0');
  if (sf_rx < 100000)
    lcd.print('0');
  if (sf_rx < 10000)
    lcd.print('0');
  if (sf_rx < 1000)
    lcd.print('0');
  if (sf_rx < 100)
    lcd.print('0');
  if (sf_rx < 10)
    lcd.print('0');
  lcd.print(sf_rx);
}
//---------------------------------------------------------------------------------------------------------
void show_inc() {
  lcd.setCursor(5, 1);
  lcd.print("        ");
  lcd.setCursor((uint8_t)(12 - log10(enc_inc)), 1);
  lcd.print('^');
}
//---------------------------------------------------------------------------------------------------------
void show_error_message() {                            // надпис при погрешно прочетени данни от ext EEPROM
  lcd.setCursor(0, 0);
  lcd.print(" Error in 24C02 ");
  delay(5000);
  lcd.clear();
}
//---------------------------------------------------------------------------------------------------------
uint32_t eeprom_read_freq(uint8_t m_addr) {                                // клетки (MSB) 1, 2, 3, 4 (LSB)
  uint32_t m_rx  = (uint32_t)eeprom.read(m_addr) << 24 |
                   (uint32_t)eeprom.read(1 + m_addr) << 16 |
                   (uint32_t)eeprom.read(2 + m_addr) << 8 |
                   (uint32_t)eeprom.read(3 + m_addr);
  return m_rx;
}
//---------------------------------------------------------------------------------------------------------
void eeprom_write_freq(uint8_t m_addr, uint32_t m_data) {                  // клетки (MSB) 1, 2, 3, 4 (LSB)
  eeprom.update(m_addr, (uint8_t)(m_data >> 24));
  eeprom.update(1 + m_addr, (uint8_t)(m_data >> 16));
  eeprom.update(2 + m_addr, (uint8_t)(m_data >> 8));
  eeprom.update(3 + m_addr, (uint8_t)m_data);
}
//---------------------------------------------------------------------------------------------------------
bool i2c_device_check(uint8_t m_addr) {                  // проверка на I2C адреса за наличие на устройство
  Wire.beginTransmission(m_addr);
  uint8_t i2c_error = Wire.endTransmission();
  if (i2c_error == 0)
    return true;
  else
    return false;
}
//---------------------------------------------------------------------------------------------------------
uint8_t button_state(uint8_t pin_bt) {       // бутон: 0(много бързо), 1(нормално), 2(бавно и при задържане)
  uint8_t bt_state;
  uint32_t t_start;
  uint32_t t_bt = 0;
  uint32_t wzab;

  if (rep)
    wzab = 100;                                  // време за една стъпка при задържане 1 сек
  else
    wzab = 1000;                                 // време за една нормална стъпка

  t_start = millis();
  while (!digitalRead(pin_bt)) {
    t_bt = millis() - t_start;
    if (t_bt > wzab) {
      rep = true;
      break;
    }
  }
  if (t_bt < 20)                                 // много бързо, без реакция (за debounce)
    bt_state = 0;
  else if (t_bt < 120)                           // нормално
    bt_state = 1;
  else                                           // бавно и при задържане
    bt_state = 2;
  return bt_state;
}
